#! /usr/bin/perl
#! c:\apl\calcula-saldos\perl5\ntt\perl.exe

# Sistema para definicao e expansao de macros.
# licensa: GNU-GPL 2
# autor: Leandro Bernsmller

$nome_prog = "mmacros.pl";
$versao_prog = "0.0.2";

# funcoes do programa
sub proc_arquivos;
sub proc_arq_macros;
sub def_macro;
sub expande_macros;
sub testa_nome_arq;
sub exibe_ambiente;

# verifica se recebeu o numero certo de parametros pela linha de comando (-1)
if ( not $#ARGV eq 0 ) {
  print "$nome_prog: erro: parmetros invlidos\n";
  print "especifique o arquivo com a enumeracao de arquivos a processar\n";
  print "use: $nome_prog arquivo-de-dados\n";
  print "exemplo: $nome_prog mproc.txt\n";
  exit 1;
}

# define nome do arquivo de dados que serah processado
$nome_cam_arq_dados = $ARGV[0];
# remove {} do nome, se houverem
$nome_cam_arq_dados =~ s/\{//i;
$nome_cam_arq_dados =~ s/\}//i;

# separa o nome e o caminho para o arquivo
$nome_arq = $nome_cam_arq_dados;
$nome_arq =~ s/.*\///i;
$caminho_arq = $&;

# verifica se arquivo existe
testa_nome_arq $caminho_arq, $nome_arq;

# inicia processamento

proc_arquivos $caminho_arq, $nome_arq;



# Processa os arquivos especificados.
# - caminho dos arquivos
# - nome do arquivo de processamento
sub proc_arquivos {

  my $caminho_arq = @_[0];  
  my $nome_arq = @_[1];

  # processa o arquivo
  open (arq_proc, "$caminho_arq$nome_arq");

  # l o arquivo, linha por linha
  while (<arq_proc>) {

    # le uma linha do arquivo
    my $op;
    my @resto;
    my @linha;
    ($op, @resto) = @linha = split ( '\s+' );

    # interpreta linha
    PROC_LIN: {

      # se  apenas um comentrio
      if ( $op eq "#" ) {
        last PROC_LIN;
      }

      # se linha esta vazia
      if ( $op eq "" ) {
        last PROC_LIN;
      }

      # processa o conteudo de cada um dos arquivos da linha
      my %amb_macros;   # define o ambiente para as macros
      foreach $nome_arq_macro (@linha) {
        testa_nome_arq $caminho_arq, $nome_arq_macro;
        %amb_macros = proc_arq_macros $caminho_arq, $nome_arq_macro, %amb_macros;
      }

      #exibe_ambiente %amb_macros;

    }   # PROC_LIN
  }     # while
}       # sub



# Processa um arquivo de macros.
# - caminho dos arquivos
# - nome do arquivo de processamento
# - ambiente com as macros
# retorna: ambiente atualizado
sub proc_arq_macros {

  my $caminho_arq = shift @_;
  my $nome_arq = shift @_;
  my %amb_macros = @_;

  # processa o arquivo
  open (arq_macro, "$caminho_arq$nome_arq");

  # l o arquivo, linha por linha
  my $estado = 0;
  my $rec = 0;
  my $nome_macro;       # variaveis para operacoes sobre macros
  my $conteudo_macro;
  my $nome_gravar;      # variaveis para gravar arquivos
  my $conteudo_gravar;
  while (<arq_macro>) {

    # converte dados lidos (ver outros chomp)
    #chomp;

    $rec = 0;

    # le uma linha do arquivo
    my $el;
    my @resto;
    my $linha;
    $linha = $_;
    ($el, @resto) = split ( '\s+' );

    # interpreta linha, por uma maquina de estados
    PROC_LIN: {
    
      if ( $estado == 0 ) {

        # se  apenas um comentrio
        if ( $el eq "#" ) {
          $rec = 1;
          last PROC_LIN;
        }

        # se linha esta vazia
        if ( $el eq "" ) {
          $rec = 1;
          last PROC_LIN;
        }

        # se linha possui operador de atribuicao
        if ( $el eq ":=" ) {
          $rec = 1;
          $estado = 1;
          last PROC_LIN;
        }

        # se linha possui operador de leitura de arquivo
        if ( $el eq ":<" ) {
          $rec = 1;
          $estado = 3;
          last PROC_LIN;
        }

        # se linha possui operador de gravacao de arquivo
        if ( $el eq ":>" ) {
          $rec = 1;
          $estado = 5;
          last PROC_LIN;
        }
        
        last PROC_LIN;
        
      } # estado 0
      
      # atribuicao - define o nome da macro
      if ( $estado == 1 ) {
        
        $nome_macro = $el;
        $conteudo_macro = "";
        $rec = 1;
        $estado = 2;
        last PROC_LIN;
      }

      # atribuicao - define conteudo da macro
      if ( $estado == 2 ) {

        # se chegou ao final do conteudo, atualiza ambiente
        if ( $el eq ";" ) {
          %amb_macros = def_macro $nome_macro, $conteudo_macro, %amb_macros;
          $rec = 1;
          $estado = 0;
          last PROC_LIN;
        }

        $conteudo_macro = join ("", $conteudo_macro, $linha);
        $rec = 1;
        last PROC_LIN;
      }

      # leitura de arquivo - define nome da macro
      if ( $estado == 3 ) {

        $nome_macro = $el;
        $conteudo_macro = "";
        $rec = 1;
        $estado = 4;
        last PROC_LIN;
      }

      # leitura de arquivo - define conteudo da macro
      if ( $estado == 4 ) {

        # se chegou ao final do conteudo, atualiza ambiente
        if ( $el eq ";" ) {
          %amb_macros = def_macro $nome_macro, $conteudo_macro, %amb_macros;
          $rec = 1;
          $estado = 0;
          last PROC_LIN;
        }

        # adiciona conteudo do arquivo especificado ao conteudo da macro
        testa_nome_arq $caminho_arq, $el;
        open (h_cont, "$caminho_arq$el");
        # le arquivo
        my $comp_arq = -s h_cont;
        my $cont_arq;
        my $num = read h_cont, $cont_arq, $comp_arq;
        chomp $cont_arq;
        $conteudo_macro = join ("", $conteudo_macro, $cont_arq);
        close h_cont;
        $rec = 1;
        last PROC_LIN;
      }

      # gravacao de arquivo - define nome do arquivo
      if ( $estado == 5 ) {

        $nome_gravar = expande_macros $el, %amb_macros;
        $conteudo_gravar = "";
        $rec = 1;
        $estado = 6;
        last PROC_LIN;
      }

      # gravacao de arquivo - define conteudo do arquivo
      if ( $estado == 6 ) {

        # se chegou ao final do conteudo, grava
        if ( $el eq ";" ) {
          open h_gr, ">$caminho_arq$nome_gravar";
          print h_gr $conteudo_gravar;
          close h_gr;
          $rec = 1;
          $estado = 0;
          last PROC_LIN;
        }

        # adiciona conteudo do arquivo especificado ao conteudo da macro
        my $cont_exp = expande_macros $el, %amb_macros;
        $conteudo_gravar = join ( "", $conteudo_gravar, $cont_exp );
        $rec = 1;
        last PROC_LIN;
      }

    }  # PROC_LIN

    print "$nome_prog: aviso: nao reconhecido: $el\n" if $rec == 0;

  }    # while

  return %amb_macros;
}



# Define uma nova macro no ambiente.
# Se jah existir, a substitui.
# - nome da macro
# - conteudo
# - ambiente (hash)
# retorna: ambiente atualizado
sub def_macro {

  my $nome_macro = shift @_;
  my $conteudo_macro = shift @_;
  my %amb_macros = @_;

  # insere no ambiente
  my $nome_exp = expande_macros $nome_macro, %amb_macros;
  # converte dados lidos (ver outros chomp)
  chomp $conteudo_macro;
  $amb_macros{$nome_exp} = $conteudo_macro;

  # tenta expandir todas as macros referidas nos conteudos do ambiente
  my $nm;  
  foreach $nm (keys %amb_macros) {
    my $cm = $amb_macros{$nm};
    $amb_macros{$nm} = expande_macros $cm, %amb_macros;
  }

  return %amb_macros;
}



# Expande as macros que forem encontradas no conteudo.
# fornecido.
# - conteudo onde procurar referencias e expandir
# - ambiente
# retorna: conteudo com as referencias expandidas
sub expande_macros {

  my $conteudo = shift @_;
  my %amb_macros = @_;

  # expande referencias no conteudo
  # verifica rapidamente se encontrou alguma referencia
  my $res = ( $conteudo =~ m/\*:.*:\*/ );
  if ($res ne "") {  # se encontrou

    # procura as referencias no conteudo
    my $pos_ini = 0;     # posicao onde inicia uma referencia
    my $pos_fim = 0;     # posicao onde termina uma referencia
    my $pos_pr = 0;      # posicao a partir de onde procura algum texto
    my $pos_fim_exp = 0; # posicao final da ultima referencia encontrada
    my $cont_exp = "";   # conteudo com referencias expandidas
    my $nome_ref = "";   # nome da macro de uma referencia
    while ($pos_ini >= 0) {

      # procura inicio de referencia
      $pos_ini = index $conteudo, "*:", $pos_pr;

      # se encontrou alguma
      if ( $pos_ini >= 0 ) {

        # procura fim da referencia
        $pos_pr = $pos_ini + 2;   # comprimento da referencia
        $pos_fim = index $conteudo, ":*", $pos_pr;

        # se encontrou fim de referencia
        if ( $pos_fim >= 0 ) {

          $pos_pr = $pos_fim + 2;

          # obtem nome da macro
          my $nome_ref = substr $conteudo, $pos_ini + 2, $pos_fim - $pos_ini - 2;

          # verifica se o nome esta em uso no ambiente
          # para cada nome no ambiente, compara
          my $nmacs;
          my $nome_encontrado = 0;
          foreach $nmacs (keys %amb_macros) {

            # se o nome estiver no ambiente, expande
            if ($nmacs eq $nome_ref) {
            # concatena macro ao conteudo resultante da expansao
              $cont_exp = join (
                "",
                $cont_exp,
                substr ($conteudo, $pos_fim_exp, $pos_ini - $pos_fim_exp),
                $amb_macros{$nome_ref}
              );
              $nome_encontrado = 1;
            }
          }

          # se nome nao estiver em uso no ambiente
          if ($nome_encontrado == 0) {
            # concatena referencia ao conteudo resultante da expansao
            $cont_exp = join (
              "",
              $cont_exp,
              substr ($conteudo, $pos_fim_exp, $pos_ini - $pos_fim_exp),
              "*:$nome_ref:*"
            );
          }

          # posicao seguinte ao final da referencia
          $pos_fim_exp = $pos_pr;
        }
      } else {

        # concatena final da ultima referencia com fim do conteudo
        $cont_exp = join (
          "",
          $cont_exp,
          substr ($conteudo, $pos_fim_exp, (length $conteudo) - $pos_fim_exp)
        );
        $pos_fim_exp = length $conteudo;
      }
    }

    # retorna conteudo com referencias expandidas
    return $cont_exp;
  }

  # retorna conteudo
  return $conteudo;
}



# Verifica se o arquivo especificado existe.
# Caso nao exista, aborta.
# - caminho onde o arquivo estah
# - nome do arquivo a testar
sub testa_nome_arq {

  my $caminho_arq = @_[0];
  my $nome_arq = @_[1];

  # verifica se arquivo nao existe
  if ( ! - -e "$caminho_arq$nome_arq" ) {
    print "$nome_prog: erro: arquivo no encontrado: $caminho_arq$nome_arq\n";
    exit 1;
  }
}



# Mostra ambiente
# - ambiente
sub exibe_ambiente {

  my %amb_macros = @_;

  print "Ambiente:\n";

  my $nmacs;
  # para cada nome no ambiente
  foreach $nmacs (keys %amb_macros) {

    print "$nmacs = $amb_macros{$nmacs}\n";
  }
}
